import os
import json
import logging
import jaydebeapi
import jpype

from ruxit.api.exceptions import AuthException, ConfigException

from segestre import Segestre, CommandLineApp
from typing import Dict, List, Tuple, Any

# TODO:
# 3. Alerts?
# 4. Metrics from DB Insights
# 5. States for tablespace
# 6. Multi dimension views of metrics (aka charts)
# 13. Figure out better way to report availability of DB vs errors in ncalculation
# 14. Control metrics from configuration page?

# DONE:
# 12. Group name is host now
# 2. change absolute counters to relative
# 11. Help texts added to every metric chart.
# 10. Lock wait time normalized to 1000 transactions.
# 7. Name metrics properly ?wtf
# 9. Proper cache usage percentage calculation

logger = logging.getLogger(__name__)

class RemoteDB2Plugin(Segestre):
    def __init__(self, *args, **kwargs):
        logger.info("Initializing RemoteDB2Plugin")

        # Obtain settings from JSON file in the same dir
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), "settings.json"))
        with open(path, mode="r", encoding="utf-8") as f:
            logger.info(f"Reading settings from {path}")
            settings_data = f.read()
            settings = json.loads(settings_data)

        # Initialize segestre and remote plugin base classes
        super().__init__(*args, **kwargs, settings=settings)

        # Boot JVM that will execute queries
        self._start_jvm()
        
    def collect(self) -> None:
        logger.info("Collecting raw metrics")
        # Collect result sets for each query from database
        queries = [
            (q, self.settings["queries"][q])
            for q in sorted(self.settings["queries"].keys())
        ]

        executed_queries = self._execute_queries(queries)
       
        # For each executed query
        for query_name, column_names, rs in executed_queries:
            dimension_keys = self.settings["queries"][query_name]["dimensions"]

            for row in rs:
                # Extract raw values from row into a dictionary in form of
                #    { column_name: value, ... }
                mapping = {str(c).lower(): row[i] for i, c in enumerate(column_names)}
                self.store(query_name, dimension_keys, mapping)
        
    def build_group_and_device_name(self) -> str:
        """Builds unique name to identify custom device"""
        c = self.settings["connection"]

        group_name = f"{c['database_host']}:{c['database_port']}"
        device_name = f"{c['database_name']}"
        return group_name, device_name

    def _build_jdbc_connection_params(self):
        c = self.settings["connection"]

        # Build connection parameters
        classname = c["jclassname"]
        url = f"{c['database_driver']}://{c['database_host']}:{c['database_port']}/{c['database_name']}"
        if c["use_ssl"]:
            url += ":sslConnection=true;"

        driver_args = [c["database_user"], c["database_password"]]

        # Log obtainer connection parameters
        logger.info(f"Connection class name: {classname}")
        logger.info(f"Connection url: {url}")
        logger.info(f"Connection args: [{driver_args[0]}, '********']")

        return classname, url, driver_args

    def _execute_queries(self, queries: List[Tuple[str, Any]]) -> None:
        logger.info("Executing queries")
        connection = cursor = None

        try:
            connection = jaydebeapi.connect(*self._build_jdbc_connection_params())
            logger.info("Connection created")
            cursor = connection.cursor()
            logger.info("Cursor obtained from connection")

            member_scope = -1 if self.settings["connection"]["single_member"] else -2

            for key, q in queries:
                if q["enabled"]:
                    logger.info(f"Executing query: {key}")
                    sql = q["sql"].format(member_scope=member_scope)
                    cursor.execute(sql)

                    rs = cursor.fetchall()
                    logger.info(f"Rowcount after execution: {cursor.rowcount}")
                    # https://www.python.org/dev/peps/pep-0249/#description
                    column_names = [c[0] for c in cursor.description]

                    yield key, column_names, rs

                else:
                    logger.info(f"Skipping query: {key}")

        except Exception as e:
            logger.exception(e)
            raise e

        finally:
            if cursor:
                cursor.close()
            if connection:
                connection.close()

    def _start_jvm(self) -> None:
        jdbc_jar_name = self.settings["connection"]["jarpath"]
        jdbc_jar_path = os.path.abspath(os.path.join(os.path.dirname(__file__), jdbc_jar_name))
        logger.info(f"Using jdbc jar: {jdbc_jar_path}")

        # Check that JVM exists somewhere. Otherwise we won't be able to
        # run JayDeBeApi
        java_home = os.environ.get("JAVA_HOME", None)
        if not java_home and not self._find_jvm_path():
            raise ConfigException("Unable to find JVM required for JDBC to run")

        # If JAVA_HOME is set, then try to find JVM in standard JAVA_HOME locations.
        # Otherwise search in predefined paths.
        jvm_path = jpype.getDefaultJVMPath() if java_home else self._find_jvm_path()
        jvm_args = f"-Djava.class.path={jdbc_jar_path}"
        logger.info(f"Found JVM at: {jvm_path}")

        try:
            logger.info(f"Attempting to start JVM at: {jvm_path}")
            jpype.startJVM(jvm_path, jvm_args, convertStrings=True)

        except OSError as ose:
            # Do not react to errors of JVM is already started for Jpype
            logger.exception(f"Ignoring OSError exception while starting JVM: {ose}")
            logger.error(f"Details: strerror: {ose.strerror}, args: {ose.args}, fn1: {ose.filename}, fn2: {ose.filename2}, errno: {ose.errno}")
            #raise ConfigException(f"Could not start JVM: {ose}")

        except Exception as e:
            logger.exception(f"Unknown exception while trying to start JVM: {e}")
            raise ConfigException(f"Could not start JVM: {e}")

    def _find_jvm_path(self):
        paths = [
            "/usr/lib/jvm",
            "../../../gateway/jre/bin/server/jvm.dll",
            "../../../gateway/jre/lib/server/libjvm.so",
            "../../../gateway/jre/lib/amd64/server/libjvm.so",
            "../../../jre/bin/server/jvm.dll",
            "../../../jre/lib/amd64/server/libjvm.so",
            "C:/Program Files/Dynatrace/gateway/jre/bin/server/jvm.dll",
            "/opt/dynatrace/gateway/jre/lib/server/libjvm.so",
            "/opt/dynatrace/gateway/jre/lib/amd64/server/libjvm.so"
        ]

        gateway_home = os.environ.get("GATEWAY_HOME", None)
        if gateway_home:
            paths.append(f"{gateway_home}/jre/bin/server/jvm.dll")
            paths.append(f"{gateway_home}/jre/bin/server/libjvm.so")

        for p in paths:
            if os.path.isfile(p):
                return p

        return None


def cli():
    logging.basicConfig(level=logging.DEBUG)
    app = CommandLineApp(RemoteDB2Plugin(results_builder=None, topology_builder=None))
    app.run()


if __name__ == "__main__":
    cli()
